from setuptools import setup, find_packages

setup(
    name = 'stClinic',
    version = '0.0.1',
    keywords='TME',
    description = 'stClinic for dissecting clinically relevant niches by integrating spatial multi-slice multi-omics data',
    license = 'MIT License',
    url = 'https://github.com/cmzuo11/stClinic',
    author = 'cmzuo',
    author_email = 'cmzuo13@163.com',
    packages = find_packages(),
    include_package_data = True,
    platforms = 'any',
    install_requires = [
        'anndata==0.9.2',
        'numpy==1.22.3',
        'igraph==0.10.8',
        'torch==1.12.0',
        'torch-cluster==1.6.1',
        'torch-geometric==2.3.0',
        'torch-scatter==2.1.1',
        'torch-sparse==0.6.17',
        'torch-spline-conv==1.2.2',
        'torchaudio==0.12.0',
        'torchvision==0.13.0',
        'tqdm==4.65.0',
        'umap-learn==0.5.3',
        'h5py==3.9.0',
        'hnswlib==0.5.1',
        'rpy2==3.5.1',
        'scanpy==1.9.3',
        'scikit-learn==1.3.0',
        'scikit-misc==0.2.0',
        'scipy==1.10.1',
        'seaborn==0.11.2',
        'lifelines==0.27.8',
        'llvmlite==0.40.1',
        'louvain==0.8.1',
        'matplotlib==3.7.2',
        'setuptools==68.0.0',
        'networkx==3.1',
        'psutil==5.9.5',
        'pycparser==2.21',
        'pyg-lib==0.2.0+pt112cu113',
        'pynndescent==0.5.10',
        'pyparsing==3.0.9',
        'PySocks==1.7.1',
        'python-dateutil==2.8.2',
        'python-igraph==0.10.8',
        'pip==23.1.2',
        ],
        )
